package com.gic.demo.common.utils;

import com.gic.redis.RedisFactory;
import org.redisson.api.RAtomicLong;
import org.redisson.api.RedissonClient;

import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.CountDownLatch;

public class UniqueIdUtils {

    private static class SnowFlake {

        private static final SnowFlake SNOW_FLAKE = new SnowFlake();

        private static final long START_TIMESTAMP = 1557489395327L; //起始的时间戳

        private static final long SEQUENCE_BIT = 12; //序列号占用位数

        private static final long MACHINE_BIT = 10; //机器标识占用位数

        private static final long TIMESTAMP_LEFT = SEQUENCE_BIT + MACHINE_BIT; //时间戳位移位数

        private static final long MAX_SEQUENCE = ~(-1L << SEQUENCE_BIT); //最大序列号

        private static final long MAX_MACHINE_ID = ~(-1L << MACHINE_BIT); //最大机器编号

        private long machineIdPart; //生成id机器标识部分

        private long sequence = 0L; //序列号

        private long lastStamp = -1L; //上一次时间戳

        private SnowFlake() {
            RedissonClient client = RedisFactory.getDefault();
            RAtomicLong atomicLong = client.getAtomicLong("SnowFlake:MachineId");
            machineIdPart = (atomicLong.addAndGet(1L) & MAX_MACHINE_ID) << SEQUENCE_BIT;
        }

        public synchronized long nextId() {
            long currentStamp = System.currentTimeMillis();
            while (currentStamp < lastStamp) { //避免机器时钟回拨
                currentStamp = System.currentTimeMillis();
            }
            if (currentStamp == lastStamp) {
                sequence++; //相同毫秒内，序列号自增
                if (sequence > MAX_SEQUENCE) { //同一毫秒的序列数已经达到最大
                    currentStamp = getNextMill();
                    sequence = 0L;
                }
            } else {
                sequence = 0L; //不同毫秒内，序列号置0
            }
            lastStamp = currentStamp;
            return (currentStamp - START_TIMESTAMP) << TIMESTAMP_LEFT | machineIdPart | sequence; //时间戳部分+机器标识部分+序列号部分
        }

        private long getNextMill() {
            long mill = System.currentTimeMillis();
            while (mill <= lastStamp) {
                mill = System.currentTimeMillis();
            }
            return mill;
        }
    }

    public static long uniqueLong() {
        return SnowFlake.SNOW_FLAKE.nextId();
    }

    public static String uniqueLongHex() {
        return String.format("%016x", uniqueLong());
    }

    public static void main(String[] args) {
        long start = System.currentTimeMillis();
        final CountDownLatch latch = new CountDownLatch(100);
        final Map<Long, Integer> map = new ConcurrentHashMap<Long, Integer>();
        for (int i = 0; i < 100; i++) {
            new Thread(new Runnable() {
                public void run() {
                    for (int i = 0; i < 20000; i++) {
                        Integer put = map.put(UniqueIdUtils.uniqueLong(), 1);
                        if (put != null) {
                            throw new RuntimeException("主键重复");
                        }
                    }
                    latch.countDown();
                }
            }).start();
        }
        try {
            latch.await();
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
        System.out.println(System.currentTimeMillis() - start);
    }
}
